#if !defined(_APIP_CHAR_P_ARITHM_H_)
#define _APIP_CHAR_P_ARITHM_H_

#include "apip_config.h"

#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#include "funcs.h"
#include "commons.h"
#include "stats.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// This symbol are not really needed but they add some consistency to the
// set of arithmetic functions.
//
// WARNING: Notice the design flaw that should be corrected: these functions
//          are implemented in commons.c.
//
#define FPM_MUL(X, Y, P)            _FpM_mul(X, Y, P)
#define FPXQ_POW(X, N, T, P)        _FpXQ_pow(X, N, T, P)
#define FPM_FPC_MUL(X, Y, P)        _FpM_FpC_mul(X, Y, P)
#define FPXQ_ISSQUARE(X, T, P)      _FpXQ_issquare(X, T, P)

#if USE_KARATSUBA == 1
    #define POL_MUL(X, Y)           RgX_mul(X, Y)
    #define POL_SQR(X)              RgX_sqr(X)
#else
    #define POL_MUL(X, Y)           ZX_mul(X, Y)
    #define POL_SQR(X)              ZX_sqr(X)
#endif

#if SHOULD_PRINT_STATS != 0
    //
    // Wrappers to basic arithmetic functions + operation counting.
    // The operation counters are defined in src/stats.h
    //
    #define SUBII(X, Y)             stat_subii(X, Y)
    #define ADDII(X, Y)             stat_addii(X, Y)
    #define ADDIS(X, Y)             stat_addis(X, Y)
    #define MULUI(X, Y)             stat_mului(X, Y)
    #define MULII(X, Y)             stat_mulii(X, Y)
    #define SQRI(X)                 stat_sqri(X)
    #define NEGI(X)                 stat_negi(X)
    #define DIVIUEXACT(X, Y)        stat_diviuexact(X, Y)
    
    #define FP_NEG(X, P)            stat_Fp_neg(X, P)
    #define FP_SUB(X, Y, P)         stat_Fp_sub(X, Y, P)
    #define FP_SQR(X, P)            stat_Fp_sqr(X, P)
    #define FP_MUL(X, Y, P)         stat_Fp_mul(X, Y, P)
    #define FP_DIV(X, Y, P)         stat_Fp_div(X, Y, P)
    #define REM(X, P)               stat_remii(X, P)
    
    #define ZX_ADD(X, Y)            stat_ZX_add(X, Y)
    #define ZX_SUB(X, Y)            stat_ZX_sub(X, Y)
    #define ZX_Z_ADD(X, Y)          stat_ZX_Z_add(X, Y)
    #define ZX_SQR(X)               stat_ZX_sqr(X)
    #define ZX_MUL(X, Y)            stat_ZX_mul(X, Y)
    #define ZX_MULS(X, Y)           stat__ZX_muls(X, Y)
    
    #define FPX_NEG(X, P)           stat_FpX_neg(X, P)
    #define FPX_ADD(X, Y, P)        stat_FpX_add(X, Y, P)
    #define FPX_SUB(X, Y, P)        stat_FpX_sub(X, Y, P)
    #define FPX_MUL(X, Y, P)        stat_FpX_mul(X, Y, P)
    #define FPX_RED(X, P)           FpX_red(X, P)
    #define FPX_REM(X, T, P)        stat_FpX_rem(X, T, P)
    #define FPX_FP_ADD(X, Y, P)     stat_FpX_Fp_add(X, Y, P)
    #define FPX_FP_MUL(X, Y, P)     stat_FpX_Fp_mul(X, Y, P)
    
    #define FPXQ_ADD(X, Y, T, P)    stat_FpXQ_add(X, Y, T, P)
    #define FPXQ_MUL(X, Y, T, P)    stat_FpXQ_mul(X, Y, T, P)
    #define FPXQ_SQR(X, T, P)       stat_FpXQ_sqr(X, T, P)
    #define FPXQ_DIV(X, Y, T, P)    stat_FpXQ_div(X, Y, T, P)
    #define FPXQ_INV(X, T, P)       stat_FpXQ_inv(X, T, P)
    
    #define FPM_FPXQ_MUL(M, X, T, P) stat_FpM_FpXQ_mul(M, X, T, P)
    
#else
    //
    // No statistics required: keep the original arithmetic functions.
    //
    #define PRINT_STATS(MSG)        /* Intentionally left empty */
    #define RESET_STATS()           /* Intentionally left empty */
    #define SET_STATS_EDEGREE(X)    /* Intentionally left empty */
    
    #define SUBII(X, Y)             subii(X, Y)
    #define ADDII(X, Y)             addii(X, Y)
    #define ADDIS(X, Y)             addis(X, Y)
    #define MULUI(X, Y)             mului(X, Y)
    #define MULII(X, Y)             mulii(X, Y)
    #define SQRI(X)                 sqri(X)
    #define NEGI(X)                 negi(X)
    #define DIVIUEXACT(X, Y)        diviuexact(X, Y)
    
    #define FP_NEG(X, P)            Fp_neg(X, P)
    #define FP_SUB(X, Y, P)         Fp_sub(X, Y, P)
    #define FP_SQR(X, P)            Fp_sqr(X, P)
    #define FP_MUL(X, Y, P)         Fp_mul(X, Y, P)
    #define FP_DIV(X, Y, P)         Fp_div(X, Y, P)
    #define REM(X, P)               remii(X, P);
                                    
    #define ZX_ADD(X, Y)            ZX_add(X, Y)
    #define ZX_SUB(X, Y)            ZX_sub(X, Y)
    #define ZX_Z_ADD(X, Y)          ZX_Z_add(X, Y)
    #define ZX_SQR(X)               POL_SQR(X)
    #define ZX_MUL(X, Y)            POL_MUL(X, Y)                             
    #define ZX_MULS(X, Y)           _ZX_muls(X, Y)
                                    
    #define FPX_NEG(X, P)           FpX_neg(X, P)
    #define FPX_ADD(X, Y, P)        FpX_add(X, Y, P)
    #define FPX_SUB(X, Y, P)        FpX_sub(X, Y, P)
    #define FPX_MUL(X, Y, P)        FpX_red(POL_MUL(X, Y), P)
    #define FPX_RED(X, P)           FpX_red(X, P)
    #define FPX_REM(X, T, P)        apip_FpX_rem(X, T, P)
    #define FPX_FP_ADD(X, Y, P)     FpX_Fp_add(X, Y, P)
    #define FPX_FP_MUL(X, Y, P)     FpX_Fp_mul(X, Y, P)
    
    #define FPXQ_ADD(X, Y, T, P)    FpXQ_add(X, Y, T, P)                             
    #define FPXQ_MUL(X, Y, T, P)    apip_FpXQ_mul(X, Y, T, P)
    #define FPXQ_SQR(X, T, P)       apip_FpXQ_sqr(X, T, P)
    #define FPXQ_DIV(X, Y, T, P)    FpXQ_div(X, Y, T, P)
    #define FPXQ_INV(X, T, P)       FpXQ_inv(X, T, P)
    
    #define FPM_FPXQ_MUL(M, X, T, P)    _FpM_FpXQ_mul(M, X, T, P)
#endif

//
// A (voluntarily bad) approximation of 1/log_10(2) (it will be used in
// conjunction with the ceil() function so a sligthly smaller approximation
// of 1/log_10(2) will be better)
//
#define INV_LOG_2        ((1/(log(2)))-0.00001)
#define N_OPERATIONS(X)  ((ulong)(ceil(log((float)(X)) * INV_LOG_2)))

//----------------------------------------------------------------------------
//
// Wrappers to basic arithmetic functions + operation counting. See the
// documentation of each wrapped function for more information. Each of these
// function follows the naming convention: stat_<name_original_function>.
//
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpM_FpXQ_mul(GEN M, GEN X, GEN T, GEN P)
{   
    //if (signe(X) != 0) {
        g2_frob_pows++;
    //}
    return _FpM_FpXQ_mul(M, X, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_remii(GEN X, GEN P) {
    return remii(X, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_rem(GEN X, GEN T, GEN P) {
    return apip_FpX_rem(X, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_ZX_mul(GEN X, GEN Y) {
    //if (signe(X) * signe(Y) != 0) {
        g2_g2_mul++;
    //}
    return POL_MUL(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat__ZX_muls(GEN X, long Y) {
    //if (signe(X) != 0) {
    #if IGNORE_MULT_BY_SMALL_CONST != 0
        g2_g2_add += N_OPERATIONS(Y);
    #else
        g1_g2_mul++;
    #endif
    //}
    return _ZX_muls(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_diviuexact(GEN X, ulong Y) {
    //
    // For the time being operations in Z are not counted. 
    //
    return diviuexact(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_negi(GEN X) {
    return negi(X);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_sqri(GEN X) {
    //if (signe(X) != 0) {
        g1_sqr++;
    //}
    return sqri(X);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_mulii(GEN X, GEN Y) {
    //if ((signe(X) != 0)) {
    //if ((signe(X) != 0) && ((signe(Y) != 0))) {
        g1_g1_mul++;
    //}
    return mulii(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_mului(ulong X, GEN Y) {
    //if (signe(Y) != 0) {
    #if IGNORE_MULT_BY_SMALL_CONST != 0
        g1_g1_add += N_OPERATIONS(X);
    #else
        g1_g1_mul++;
    #endif
    //}
    return mului(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_addis(GEN X, long Y) {
    //
    // For the time being operations in Z are not counted. 
    //
    return addis(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_addii(GEN X, GEN Y) {
    g1_g1_add++;
    return addii(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_subii(GEN X, GEN Y) {
    g1_g1_add++;
    return subii(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpXQ_inv(GEN X, GEN T, GEN P) {
    g2_inv++;
    return FpXQ_inv(X, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_ZX_Z_add(GEN X, GEN Y) {
    //if ((signe(X) != 0) && (signe(Y) != 0)) {
        g1_g2_add++;
    //}
    return ZX_Z_add(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_ZX_sub(GEN X, GEN Y) {
    //if ((signe(X) != 0) && (signe(Y) != 0)) {
        g2_g2_add++;
    //}
    return ZX_sub(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_ZX_add(GEN X, GEN Y) {
    //if ((signe(X) != 0) && (signe(Y) != 0)) {
        g2_g2_add++;
    //}
    return ZX_add(X, Y);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_ZX_sqr(GEN X) {
    //if (signe(X) != 0) {
        g2_sqr++;
    //}
    return POL_SQR(X);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpXQ_div(GEN X, GEN Y, GEN T, GEN P) {
    g2_inv++;
    g2_g2_mul++;
    return FpXQ_div(X, Y, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_Fp_mul(GEN X, GEN Y, GEN P) {
    //if (signe(X) * signe(Y) != 0) {
        g1_g2_mul++;
    //}
    return FpX_Fp_mul(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_Fp_sub(GEN X, GEN Y, GEN P) {
    g1_g1_add++;
    return Fp_sub(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_add(GEN X, GEN Y, GEN P) {
    g2_g2_add++;
    return FpX_add(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_sub(GEN X, GEN Y, GEN P) {
    g2_g2_add++;
    return FpX_sub(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_Fp_div(GEN X, GEN Y, GEN P) {
    g1_inv++;
    g1_g1_mul++;
    return Fp_div(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_neg(GEN X, GEN P) {
    return FpX_neg(X, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_mul(GEN X, GEN Y, GEN P) {
    //if (signe(X) * signe(Y) != 0) {
        g2_g2_mul++;
    //}
    return FpX_red(POL_MUL(X, Y), P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_Fp_neg(GEN X, GEN P) {
    return Fp_neg(X, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_Fp_sqr(GEN X, GEN P) {
    //if (signe(X) != 0) {
        g1_sqr++;
    //}
    return Fp_sqr(X, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_Fp_mul(GEN X, GEN Y, GEN P) {
    //if (signe(X) * signe(Y) != 0) {
        g1_g1_mul++;
    //}
    return Fp_mul(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpX_Fp_add(GEN X, GEN Y, GEN P) {
    //if (signe(X) * signe(Y) != 0) {
        g1_g2_add++;
    //}
    return FpX_Fp_add(X, Y, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpXQ_sqr(GEN X, GEN T, GEN P) {
    //if (signe(X) != 0) {
        g2_sqr++;
    //}
    return apip_FpXQ_sqr(X, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpXQ_mul(GEN X, GEN Y, GEN T, GEN P) {
    //if (signe(X) * signe(Y) != 0) {
        g2_g2_mul++;
    //}
    return apip_FpXQ_mul(X, Y, T, P);
}
//----------------------------------------------------------------------------
MAYBE_UNUSED INLINE GEN
stat_FpXQ_add(GEN X, GEN Y, GEN T, GEN P) {
    g2_g2_add++;
    return FpXQ_add(X, Y, T, P);
}
//----------------------------------------------------------------------------

#ifdef __cplusplus
}
#endif
#endif
