#include <strings.h>

#if !defined(_PARI_H_)
    #define _PARI_H_
    #include <pari.h>
#endif

#include "apip_config.h"
#include "pairings.h"
#include "funcs.h"
#include "stats.h"

#if APIP_POINTER_BITSIZE == 64
    #define LOW_32BIT(X)  ( ((unsigned long)(X)) & 0x00000000FFFFFFFF)
    #define HIGH_32BIT(X) ((((unsigned long)(X)) & 0xFFFFFFFF00000000) >> 32)
#else
    #define LOW_32BIT(X)  ((unsigned long)(X))
    #define HIGH_32BIT(X) ((unsigned long)(X))
#endif

//-----------------------------------------------------------------------------
static GEN pointer_to_gen(void* ptr) {
    pari_sp ltop = avma;
    GEN genptr = utoi((unsigned long)ptr);
    return gerepileupto(ltop, genptr);    
}
//-----------------------------------------------------------------------------
static void* gen_to_pointer(GEN genptr) {    
    ulong ptr = itou(genptr);
    return (void*)ptr;
}
//-----------------------------------------------------------------------------
GEN
apip_alloc_pairing(GEN curve, GEN charac,
                   GEN f1_defpol, GEN f2_defpol,
                   GEN g1_order, GEN sfamily)
{
    IN();
    pari_sp ltop = avma;
    
    curve_family_t family = curve_family_from_str(GSTR(sfamily));
    
    pairing_data_t*
    result = alloc_pairing_data(
                 curve, charac, f1_defpol, f2_defpol, g1_order, family
             );
    
    GEN genptr;
    if (result == NULL) {
        genptr = gen_0;
    } else {
        genptr = pointer_to_gen((void*)result);
    }
    OUT();
    return gerepileupto(ltop, genptr);
}
//-----------------------------------------------------------------------------
void
apip_free_pairing(GEN gen_pdata)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    free_pairing_data(pdata);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_coord(GEN gen_pdata, char* coordsys)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->coordsys = coord_system_from_str(coordsys);
    update_pairing_data(pdata);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_miller(GEN gen_pdata, char* method)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->miller_method = miller_method_from_str(method);
    update_pairing_data(pdata);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_cofac(GEN gen_pdata, GEN cof)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->cof = gclone(cof);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_hardpow_method(GEN gen_pdata, char* method)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->hardpow_method = hardpow_method_from_str(method);
    pdata->r = 0;
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_family(GEN gen_pdata, char* family)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->family = curve_family_from_str(family);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_family_param(GEN gen_pdata, GEN u)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->u = gclone(u);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_denom_elim(GEN gen_pdata, long do_denom_elim)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    if (IS_EVEN(pdata->embed_deg)) {
        if (do_denom_elim == 0) {
            pdata->do_denom_elim = 0;
        } else {
            pdata->do_denom_elim = 1;
        }
    } else {
        pariprintf(
            "Odd embedding degree -- no denominator elimination possible\n"
        );
        pariflush();
    }
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_do_reduce(GEN gen_pdata, long do_reduce)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    if (do_reduce == 0) {
        pdata->do_reduce = 0;
    } else {
       pdata->do_reduce = 1;
    }
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_do_naive_exp(GEN gen_pdata, long do_naive_exp)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    if (do_naive_exp == 0) {
        pdata->do_naive_exp = 0;
    } else {
        pdata->do_naive_exp = 1;
    }
    update_pairing_data(pdata);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_do_naive_hard_exp(GEN gen_pdata, long do_naive_hard_exp)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    if (do_naive_hard_exp == 0) {
        pdata->do_naive_hard_exp = 0;
    } else {
        pdata->do_naive_hard_exp = 1;
    }
    update_pairing_data(pdata);
    return;
}
//-----------------------------------------------------------------------------
void
apip_set_frob_trace(GEN gen_pdata, GEN t)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->frob_trace = t;
    return;
}
//-----------------------------------------------------------------------------
GEN
apip_weil(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return weil(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_tate(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return tate(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_ate(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return ate(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_twisted(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return twisted(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_opti_ate(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return opti_ate(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_opti_twisted(GEN gen_pdata, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return opti_twisted(pdata, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_miller(GEN gen_pdata, GEN r, GEN P, GEN Q)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return miller(pdata, r, P, Q);
}
//-----------------------------------------------------------------------------
GEN
apip_final_exp(GEN gen_pdata, GEN unreduced)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return pdata->final_exp_func(pdata, unreduced);
}
//-----------------------------------------------------------------------------
GEN
apip_frobenius_power(GEN gen_pdata, GEN X, long i)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return frobenius_power(pdata, X, i);
}
//-----------------------------------------------------------------------------
void
apip_compute_frobenius_powers(GEN gen_pdata)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    compute_frobenius_powers(pdata);
}
//-----------------------------------------------------------------------------
void
apip_print_frobenius_matrix(GEN gen_pdata, long i)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    print_frobenius_matrix(pdata, i);
}
//-----------------------------------------------------------------------------
void
apip_print_pairing(GEN gen_pdata)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pariprintf("-------------------------------------------------------\n");
    pariprintf("         a : %s\n", GENtostr(ell_get_a4(pdata->curve)));
    pariprintf("         b : %s\n", GENtostr(ell_get_a6(pdata->curve)));
    pariprintf("    family : %s\n", curve_families_to_str[pdata->family]);
    pariprintf("    charac : %s\n", GENtostr(pdata->charac));
    if (pdata->frob_trace == NULL) {
        pariprintf("         t : unknown\n");
    } else {
        pariprintf("         t : %s\n", GENtostr(pdata->frob_trace));
    }
    pariprintf("    f1 ext : %i\n", pdata->f1_extdeg);
    pariprintf("    f2 ext : %i\n", pdata->f2_extdeg);

    if (pdata->fields_are_friendly) {
        pariprintf("  friendly : no\n");
    } else {
        pariprintf("  friendly : fields are friendly [KM05]\n");
    }
    pariprintf("   #Aut(E) : %i\n", pdata->card_aut_E);
    pariprintf("-------------------------------------------------------\n");
    pariprintf("  card(g1) : %s\n", GENtostr(pdata->g1_order));
    pariprintf("   emb deg : %i\n", pdata->embed_deg);
    pariprintf(" final exp : %s\n", GENtostr(pdata->final_exp));
    pariprintf("easy exp 1 : %s\n", GENtostr(pdata->easy_exp_1));
    pariprintf("easy exp 2 : %s\n", GENtostr(pdata->easy_exp_2));
    pariprintf("  hard exp : %s\n", GENtostr(pdata->hard_exp));
    pariprintf("-------------------------------------------------------\n");
    pariprintf("     coord : %s\n", coord_systems_to_str[pdata->coordsys]);
    pariprintf("    miller : %s\n", miller_methods_to_str[pdata->miller_method]);
    pariprintf("  hard exp : %s\n", hardpow_methods_to_str[pdata->miller_method]);
    pariprintf("-------------------------------------------------------\n");
    if (pdata->do_denom_elim) {
        pariprintf("     denom : with denominator elimination\n");
    } else {
        pariprintf("     denom : no denominator elimination\n");
    }
    if (pdata->precomp_done) {
        pariprintf("   precomp : precomputations done\n");
    } else {
        pariprintf("   precomp : no precomputations done\n");
    }
    if (pdata->do_naive_exp) {
        pariprintf("       exp : naive\n");
    } else {
        pariprintf("       exp : optimized\n");
    }
    if (pdata->do_naive_hard_exp) {
        pariprintf("  hard exp : naive\n");
    } else {
        pariprintf("  hard exp : optimized if (exp == optimized)\n");
    }
    pariprintf("-------------------------------------------------------\n\n");
    pariflush();
    return;
}
//-----------------------------------------------------------------------------
void
apip_print_avma() {
    pariprintf("AVMA = %u\n", avma);
}
//-----------------------------------------------------------------------------
void
apip_set_stats_verbosity(ulong verbosity)
{
    switch (verbosity) {
        case STATS_DETAILLED:
        case STATS_SIMPLIFIED:
        case STATS_SHORT:
        case STATS_SHORTEST:
            stats_verbosity = verbosity;
            break;
        default:
            stats_verbosity = DFLT_STATS_VERBOSITY;
    }
}
//-----------------------------------------------------------------------------
void
apip_set_stats_ratios(GEN g1_sm, GEN g2_sm, GEN g1_im, GEN g2_im)
{
    sqr_to_mul_ratio_g1 = gtodouble(g1_sm);
    sqr_to_mul_ratio_g2 = gtodouble(g2_sm);
    inv_to_mul_ratio_g1 = gtodouble(g1_im);
    inv_to_mul_ratio_g2 = gtodouble(g2_im);
}
//-----------------------------------------------------------------------------
void apip_set_r(GEN gen_pdata, long r)
{
    //
    // For debugging purpose only -- don't use me
    //
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->r = r;
    return;
}
//-----------------------------------------------------------------------------
void apip_set_twist_degree(GEN gen_pdata, long d)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    set_twist_degree(pdata, d);
    return;
}
//-----------------------------------------------------------------------------
GEN apip_get_point(GEN gen_pdata)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    return gcopy(pdata->point);
}
//-----------------------------------------------------------------------------
void apip_set_frobcost(GEN gen_pdata, GEN cost)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    // Force recomputation of optimal parameter r since cost evaluation changed
    pdata->frobcost = (float)rtodbl(cost);
    pdata->r = 0;
    return;
}
//-----------------------------------------------------------------------------
void apip_set_sqrcost(GEN gen_pdata, GEN cost)
{
    pairing_data_t* pdata = (pairing_data_t*) gen_to_pointer(gen_pdata);
    pdata->sqrcost = (float)rtodbl(cost);
    // Force recomputation of optimal parameter r since cost evaluation changed
    pdata->r = 0;
    return;
}
//-----------------------------------------------------------------------------
